#include <cstring>
#include <string>
#include <msp/core/maputils.h>
#include <msp/strings/lexicalcast.h>
#include "protocol.h"

using namespace std;


namespace Msp {
namespace Net {

Protocol::Protocol(unsigned npi):
	header_def(0),
	next_packet_id(npi)
{
	PacketDefBuilder<PacketHeader, NullSerializer<PacketHeader> >(*this, header_def, NullSerializer<PacketHeader>())
		(&PacketHeader::type)(&PacketHeader::length);
}

Protocol::~Protocol()
{
	for(map<unsigned, PacketDefBase *>::iterator i=packet_class_defs.begin(); i!=packet_class_defs.end(); ++i)
		delete i->second;
}

void Protocol::add_packet(PacketDefBase *pdef)
{
	PacketDefBase *&ptr = packet_class_defs[pdef->get_class_id()];
	if(ptr)
		delete ptr;
	ptr = pdef;
	packet_id_defs[pdef->get_id()] = pdef;
}

const Protocol::PacketDefBase &Protocol::get_packet_by_class_id(unsigned id) const
{
	return *get_item(packet_class_defs, id);
}

const Protocol::PacketDefBase &Protocol::get_packet_by_id(unsigned id) const
{
	return *get_item(packet_id_defs, id);
}

unsigned Protocol::dispatch(ReceiverBase &rcv, const char *buf, unsigned size) const
{
	PacketHeader header;
	buf = header_def.deserialize(header, buf, buf+size);
	if(header.length>size)
		throw bad_packet("truncated");
	const PacketDefBase &pdef = get_packet_by_id(header.type);
	const char *ptr = pdef.dispatch(rcv, buf, buf+header.length);
	return ptr-buf;
}

unsigned Protocol::get_packet_size(const char *buf, unsigned size) const
{
	if(size<4)
		return 0;
	PacketHeader header;
	header_def.deserialize(header, buf, buf+size);
	return header.length;
}

unsigned Protocol::get_hash() const
{
	// TODO
	return 123;
}


/* TODO These assumes the machine is little-endian; are there any relevant
big-endian platforms these days? */
template<typename T>
char *Protocol::BasicSerializer<T>::serialize(const T &value, char *buf, char *end) const
{
	if(end-buf<static_cast<int>(sizeof(T)))
		throw buffer_error("overflow");

	const char *ptr = reinterpret_cast<const char *>(&value)+sizeof(T);
	for(unsigned i=0; i<sizeof(T); ++i)
		*buf++ = *--ptr;

	return buf;
}

template<typename T>
const char *Protocol::BasicSerializer<T>::deserialize(T &value, const char *buf, const char *end) const
{
	if(end-buf<static_cast<int>(sizeof(T)))
		throw buffer_error("underflow");

	char *ptr = reinterpret_cast<char *>(&value)+sizeof(T);
	for(unsigned i=0; i<sizeof(T); ++i)
		*--ptr = *buf++;

	return buf;
}

template char *Protocol::BasicSerializer<Int8>::serialize(const Int8 &, char *, char *) const;
template char *Protocol::BasicSerializer<Int16>::serialize(const Int16 &, char *, char *) const;
template char *Protocol::BasicSerializer<Int32>::serialize(const Int32 &, char *, char *) const;
template char *Protocol::BasicSerializer<Int64>::serialize(const Int64 &, char *, char *) const;
template char *Protocol::BasicSerializer<UInt8>::serialize(const UInt8 &, char *, char *) const;
template char *Protocol::BasicSerializer<UInt16>::serialize(const UInt16 &, char *, char *) const;
template char *Protocol::BasicSerializer<UInt32>::serialize(const UInt32 &, char *, char *) const;
template char *Protocol::BasicSerializer<UInt64>::serialize(const UInt64 &, char *, char *) const;
template char *Protocol::BasicSerializer<float>::serialize(const float &, char *, char *) const;
template char *Protocol::BasicSerializer<double>::serialize(const double &, char *, char *) const;
template const char *Protocol::BasicSerializer<Int8>::deserialize(Int8 &, const char *, const char *) const;
template const char *Protocol::BasicSerializer<Int16>::deserialize(Int16 &, const char *, const char *) const;
template const char *Protocol::BasicSerializer<Int32>::deserialize(Int32 &, const char *, const char *) const;
template const char *Protocol::BasicSerializer<Int64>::deserialize(Int64 &, const char *, const char *) const;
template const char *Protocol::BasicSerializer<UInt8>::deserialize(UInt8 &, const char *, const char *) const;
template const char *Protocol::BasicSerializer<UInt16>::deserialize(UInt16 &, const char *, const char *) const;
template const char *Protocol::BasicSerializer<UInt32>::deserialize(UInt32 &, const char *, const char *) const;
template const char *Protocol::BasicSerializer<UInt64>::deserialize(UInt64 &, const char *, const char *) const;
template const char *Protocol::BasicSerializer<float>::deserialize(float &, const char *, const char *) const;
template const char *Protocol::BasicSerializer<double>::deserialize(double &, const char *, const char *) const;


Protocol::StringSerializer::StringSerializer(const Protocol &p):
	length_serializer(p)
{ }

char *Protocol::StringSerializer::serialize(const string &str, char *buf, char *end) const
{
	buf = length_serializer.serialize(str.size(), buf, end);
	if(end-buf<static_cast<int>(str.size()))
		throw buffer_error("overflow");
	copy(str.begin(), str.end(), buf);
	return buf+str.size();
}

const char *Protocol::StringSerializer::deserialize(string &str, const char *buf, const char *end) const
{
	UInt16 length;
	buf = length_serializer.deserialize(length, buf, end);
	if(end-buf<static_cast<int>(length))
		throw buffer_error("underflow");
	str.assign(buf, buf+length);
	return buf+length;
}


unsigned Protocol::PacketDefBase::next_class_id = 1;

Protocol::PacketDefBase::PacketDefBase(unsigned i):
	id(i)
{ }


Protocol::PacketHeader::PacketHeader():
	type(0),
	length(0)
{ }

Protocol::PacketHeader::PacketHeader(UInt16 t, UInt16 l):
	type(t),
	length(l)
{ }

} // namespace Net
} // namespace Msp
