#include "platform_api.h"
#include <msp/core/systemerror.h>
#include <msp/io/handle_private.h>
#include "sockaddr_private.h"
#include "socket.h"
#include "socket_private.h"

namespace Msp {
namespace Net {

Socket::Socket(const Private &p):
	priv(new Private),
	local_addr(0)
{
	mode = IO::M_RDWR;

	priv->handle = p.handle;

	SockAddr::SysAddr sa;
	getsockname(priv->handle, reinterpret_cast<sockaddr *>(&sa.addr), &sa.size);
	local_addr = SockAddr::new_from_sys(sa);

	platform_init();
}

Socket::Socket(Family af, int type, int proto):
	priv(new Private),
	local_addr(0)
{
	mode = IO::M_RDWR;

	priv->handle = socket(family_to_sys(af), type, proto);

	platform_init();
}

Socket::~Socket()
{
	platform_cleanup();

	delete local_addr;
	delete priv;
}

void Socket::set_block(bool b)
{
	mode = (mode&~IO::M_NONBLOCK);
	if(b)
		mode = (mode|IO::M_NONBLOCK);

	priv->set_block(b);
}

const IO::Handle &Socket::get_event_handle()
{
	return priv->event;
}

void Socket::bind(const SockAddr &addr)
{
	SockAddr::SysAddr sa = addr.to_sys();

	int err = ::bind(priv->handle, reinterpret_cast<sockaddr *>(&sa.addr), sa.size);
	if(err==-1)
		throw system_error("bind");

	delete local_addr;
	local_addr = addr.copy();
}

const SockAddr &Socket::get_local_address() const
{
	if(local_addr==0)
		throw bad_socket_state("not bound");
	return *local_addr;
}

void Socket::set_socket_events(unsigned e)
{
	IO::PollEvent pe = static_cast<IO::PollEvent>(e&4095);
	if(e&S_CONNECT)
		pe = pe|IO::P_OUTPUT;
	if(e&S_ACCEPT)
		pe = pe|IO::P_INPUT;

	set_platform_events(e);
	set_events(pe);
}

} // namespace Net
} // namespace Msp
