#include "platform_api.h"
#include <msp/core/systemerror.h>
#include <msp/io/handle_private.h>
#include <msp/io/poll.h>
#include <msp/strings/format.h>
#include "sockaddr_private.h"
#include "socket_private.h"
#include "streamsocket.h"

namespace Msp {
namespace Net {

StreamSocket::StreamSocket(const Private &p, const SockAddr &paddr):
	ClientSocket(p, paddr)
{
	set_socket_events(S_INPUT);
}

StreamSocket::StreamSocket(Family af, int proto):
	ClientSocket(af, SOCK_STREAM, proto)
{ }

bool StreamSocket::connect(const SockAddr &addr)
{
	if(connected)
		throw bad_socket_state("already connected");

	SockAddr::SysAddr sa = addr.to_sys();
	int err = ::connect(priv->handle, reinterpret_cast<const sockaddr *>(&sa.addr), sa.size);
	bool finished = check_sys_connect_error(err);
	if(!finished)
	{
		connecting = true;
		set_socket_events(S_CONNECT);
	}

	delete peer_addr;
	peer_addr = addr.copy();

	delete local_addr;
	SockAddr::SysAddr lsa;
	getsockname(priv->handle, reinterpret_cast<sockaddr *>(&lsa.addr), &lsa.size);
	local_addr = SockAddr::new_from_sys(lsa);

	if(finished)
	{
		connected = true;
		set_socket_events(S_INPUT);
		signal_connect_finished.emit(0);
	}

	return connected;
}

bool StreamSocket::poll_connect(const Time::TimeDelta &timeout)
{
	if(!connecting)
		return false;

	IO::PollEvent res = poll(*this, IO::P_OUTPUT, timeout);
	if(res&IO::P_OUTPUT)
	{
		connecting = false;

		int err;
		socklen_t len = sizeof(int);
		priv->get_option(SOL_SOCKET, SO_ERROR, &err, &len);

		if(err!=0)
		{
			set_socket_events(S_NONE);
			throw system_error("connect", err);
		}

		set_socket_events(S_INPUT);

		connected = true;
	}

	return connected;
}

void StreamSocket::on_event(IO::PollEvent ev)
{
	if((ev&(IO::P_OUTPUT|IO::P_ERROR)) && connecting)
	{
		int err;
		socklen_t len = sizeof(err);
		priv->get_option(SOL_SOCKET, SO_ERROR, &err, &len);

		connecting = false;
		connected = (err==0);
		if(err)
		{
			system_error exc("connect", err);
			signal_connect_finished.emit(&exc);
		}
		else
			signal_connect_finished.emit(0);

		if(err!=0)
		{
			delete peer_addr;
			peer_addr = 0;
		}

		set_socket_events((err==0) ? S_INPUT : S_NONE);
	}
}

} // namespace Net
} // namespace Msp
